<?php
/* vim: set expandtab tabstop=4 shiftwidth=4 foldmethod=marker: */
// +----------------------------------------------------------------------+
// | PHP Version 4                                                        |
// +----------------------------------------------------------------------+
// | Copyright (c) 1997-2004 The PHP Group                                |
// +----------------------------------------------------------------------+
// | This source file is subject to version 2.02 of the PHP license,      |
// | that is bundled with this package in the file LICENSE, and is        |
// | available at through the world-wide-web at                           |
// | http://www.php.net/license/2_02.txt.                                 |
// | If you did not receive a copy of the PHP license and are unable to   |
// | obtain it through the world-wide-web, please send a note to          |
// | license@php.net so we can mail you a copy immediately.               |
// +----------------------------------------------------------------------+
// | Author: Robert Twitty <rtwitty@users.sourceforge.net>                |
// +----------------------------------------------------------------------+
//
// $Id: odbtp.php,v 1.7 2004/05/26 00:24:18 rtwitty Exp $

require_once 'DB/common.php';

/**
 * Database independent query interface definition for ODBTP extension.
 *
 * @package  DB
 * @version  $Id: odbtp.php,v 1.7 2004/05/26 00:24:18 rtwitty Exp $
 * @category Database
 * @author   Robert Twitty <rtwitty@users.sourceforge.net>
 */
class DB_odbtp extends DB_common
{
    // {{{ properties

    var $connection;
    var $phptype, $dbsyntax;
    var $prepare_tokens = array();
    var $prepare_types = array();
    var $_db = null;
    var $odbc_driver;
    var $query_result;

    // }}}
    // {{{ constructor

    function DB_odbtp()
    {
        $this->DB_common();
        $this->phptype = 'odbtp';
        $this->dbsyntax = 'odbtp';
        $this->features = array(
            'prepare' => false,
            'pconnect' => true,
            'transactions' => false,
            'limit' => 'emulate'
        );
        $this->errorcode_map = array(
            '01004' => DB_ERROR_TRUNCATED,
            '07001' => DB_ERROR_MISMATCH,
            '21S01' => DB_ERROR_MISMATCH,
            '21S02' => DB_ERROR_MISMATCH,
            '22003' => DB_ERROR_INVALID_NUMBER,
            '22007' => DB_ERROR_INVALID_DATE,
            '22018' => DB_ERROR_INVALID_NUMBER,
            '22012' => DB_ERROR_DIVZERO,
            '23000' => DB_ERROR_CONSTRAINT,
            '23503' => DB_ERROR_CONSTRAINT,
            '24000' => DB_ERROR_INVALID,
            '34000' => DB_ERROR_INVALID,
            '42000' => DB_ERROR_SYNTAX,
            '42S01' => DB_ERROR_ALREADY_EXISTS,
            '42S02' => DB_ERROR_NOSUCHTABLE,
            '42S11' => DB_ERROR_ALREADY_EXISTS,
            '42S12' => DB_ERROR_NOT_FOUND,
            '42S21' => DB_ERROR_ALREADY_EXISTS,
            '42S22' => DB_ERROR_NOSUCHFIELD,
            '08004' => DB_ERROR_CONNECT_FAILED,
            '08007' => DB_ERROR_CONNECT_FAILED,
            '08S01' => DB_ERROR_CONNECT_FAILED,
            'HY009' => DB_ERROR_INVALID,
            'HY024' => DB_ERROR_INVALID,
            'HY090' => DB_ERROR_INVALID,
            'IM001' => DB_ERROR_UNSUPPORTED,
            'ODBTPINV' => DB_ERROR_INVALID
        );
    }

    // }}}
    // {{{ connect()

    function connect($dsninfo, $persistent = false)
    {
        if (!DB::assertExtension('odbtp')) {
            return $this->raiseError(DB_ERROR_EXTENSION_NOT_FOUND);
        }
        $this->dsn = $dsninfo;
        $user = $dsninfo['username'];
        $pw = $dsninfo['password'];
        $db = $dsninfo['database'];
        $dbhost = $dsninfo['hostspec'] ? $dsninfo['hostspec'] : 'localhost';
        $dboptions = strtolower($dsninfo['dbsyntax']);

        if ($dsninfo['port']) {
            $dbhost .= ':' . $dsninfo['port'];
        }
        if (strstr($dboptions, 'reserved')) {
            $connect_function = 'odbtp_rconnect';
        } else if (strstr($dboptions, 'single')) {
            $connect_function = 'odbtp_sconnect';
        } else {
            $connect_function = 'odbtp_connect';
        }
        $conn = @$connect_function($dbhost, $user, $pw, $db);
        if (!$conn) {
            return $this->raiseError(DB_ERROR_CONNECT_FAILED,
                                     null, null, null, @odbtp_last_error());
        }
        if( $connect_function == 'odbtp_connect' && !$persistent ) {
            @odbtp_dont_pool_dbc($conn);
        }
        if (strstr($dboptions, 'unicode')) {
            if (!@odbtp_set_attr(ODB_ATTR_UNICODESQL, 1, $conn)) {
                return $this->raiseError(DB_ERROR, null, null, null,
                                         @odbtp_last_error());
            }
        }
        if (!@odbtp_set_attr(ODB_ATTR_FULLCOLINFO, 1, $conn)) {
            return $this->raiseError(DB_ERROR, null, null, null,
                                     @odbtp_last_error());
        }
        $this->odbc_driver = @odbtp_get_attr(ODB_ATTR_DRIVER, $conn);

        switch( $this->odbc_driver ) {
            case ODB_DRIVER_JET:
            case ODB_DRIVER_FOXPRO:
                $tc = false;
                break;
            case ODB_DRIVER_MSSQL:
            case ODB_DRIVER_ORACLE:
            case ODB_DRIVER_SYBASE:
                $tc = true;
                break;
            default:
                $tc = @odbtp_get_attr(ODB_ATTR_TXNCAPABLE, $conn);
        }
        $this->features['transactions'] = $tc ? true : false;

        $this->connection = $conn;
        $this->_db = $db;
        return DB_OK;
    }

    // }}}
    // {{{ disconnect()

    function disconnect()
    {
        $ret = @odbtp_close($this->connection);
        $this->connection = null;
        return $ret;
    }

    // }}}
    // {{{ simpleQuery()

    function simpleQuery($query)
    {
        $ismanip = DB::isManip($query);
        $this->last_query = $query;
        $query = $this->modifyQuery($query);
        $this->query_result = @odbtp_query($query, $this->connection);
        if (!$this->query_result) {
            return $this->odbtpRaiseError();
        }
        // Determine which queries that should return data, and which
        // should return an error code only.
        return $ismanip ? DB_OK : $this->query_result;
    }

    // }}}

    // {{{ nextResult()

    /**
     * Move the internal odbtp result pointer to the next available result
     *
     * @param a valid fbsql result resource
     *
     * @access public
     *
     * @return true if a result is available otherwise return false
     */
    function nextResult($result)
    {
        return @odbtp_next_result($result);
    }

    // }}}
    // {{{ fetchInto()

    /**
     * Fetch a row and insert the data into an existing array.
     *
     * Formating of the array and the data therein are configurable.
     * See DB_result::fetchInto() for more information.
     *
     * @param resource $result    query result identifier
     * @param array    $arr       (reference) array where data from the row
     *                            should be placed
     * @param int      $fetchmode how the resulting array should be indexed
     * @param int      $rownum    the row number to fetch
     *
     * @return mixed DB_OK on success, NULL when end of result set is
     *               reached or on failure
     *
     * @see DB_result::fetchInto()
     * @access private
     */
    function fetchInto($result, &$arr, $fetchmode, $rownum=null)
    {
        if ($rownum !== null) {
            if (!@odbtp_data_seek($result, $rownum)) {
                return null;
            }
        }
        if ($fetchmode & DB_FETCHMODE_ASSOC) {
            $arr = @odbtp_fetch_assoc($result);
            if ($this->options['portability'] & DB_PORTABILITY_LOWERCASE && $arr) {
                $arr = array_change_key_case($arr, CASE_LOWER);
            }
        } else {
            $arr = @odbtp_fetch_row($result);
        }
        if (!$arr) {
            /* This throws informative error messages,
               don't use it for now
            if (@odbtp_get_error($result)) {
                return $this->odbtpRaiseError();
            }
            */
            return null;
        }
        if ($this->options['portability'] & DB_PORTABILITY_RTRIM) {
            $this->_rtrimArrayValues($arr);
        }
        if ($this->options['portability'] & DB_PORTABILITY_NULL_TO_EMPTY) {
            $this->_convertNullArrayValuesToEmpty($arr);
        }
        return DB_OK;
    }

    // }}}
    // {{{ freeResult()

    function freeResult($result)
    {
        return @odbtp_free_query($result);
    }

    // }}}
    // {{{ numCols()

    function numCols($result)
    {
        return @odbtp_num_fields($result);
    }

    // }}}
    // {{{ numRows()

    function numRows($result)
    {
        $rows = @odbtp_num_rows($result);
        if ($rows === false) {
            return $this->odbtpRaiseError();
        }
        return $rows;
    }

    // }}}
    // {{{ autoCommit()

    /**
     * Enable/disable automatic commits
     */
    function autoCommit($onoff = false)
    {
        $txn = $onoff ? ODB_TXN_NONE : ODB_TXN_READUNCOMMITTED;
        if (!@odbtp_set_attr(ODB_ATTR_TRANSACTIONS, $txn, $this->connection)) {
            return $this->odbtpRaiseError();
        }
        return DB_OK;
    }

    // }}}
    // {{{ commit()

    /**
     * Commit the current transaction.
     */
    function commit()
    {
        if (!@odbtp_commit($this->connection)) {
            return $this->odbtpRaiseError();
        }
        return DB_OK;
    }

    // }}}
    // {{{ rollback()

    /**
     * Roll back (undo) the current transaction.
     */
    function rollback()
    {
        if (!@odbtp_rollback($this->connection)) {
            return $this->odbtpRaiseError();
        }
        return DB_OK;
    }

    // }}}
    // {{{ affectedRows()

    /**
     * Gets the number of rows affected by the last query.
     * if the last query was a select, returns 0.
     *
     * @return number of rows affected by the last query or DB_ERROR
     */
    function affectedRows()
    {
        if (DB::isManip($this->last_query)) {
            $result = @odbtp_affected_rows($this->query_result);
        } else {
            $result = 0;
        }
        return $result;
    }

    // }}}
    // {{{ nextId()

    /**
     * Returns the next free id in a sequence
     *
     * @param string  $seq_name  name of the sequence
     * @param boolean $ondemand  when true, the seqence is automatically
     *                           created if it does not exist
     *
     * @return int  the next id number in the sequence.  DB_Error if problem.
     *
     * @internal
     * @see DB_common::nextID()
     * @access public
     */
    function nextId($seq_name, $ondemand = true)
    {
        $seqname = $this->getSequenceName($seq_name);
        $repeat = 0;
        do {
            $this->pushErrorHandling(PEAR_ERROR_RETURN);
            $result = $this->query("UPDATE ${seqname} SET id = id + 1");
            $this->popErrorHandling();
            if ($ondemand && DB::isError($result) &&
                $result->getCode() == DB_ERROR_NOSUCHTABLE) {
                $repeat = 1;
                $this->pushErrorHandling(PEAR_ERROR_RETURN);
                $result = $this->createSequence($seq_name);
                $this->popErrorHandling();
                if (DB::isError($result)) {
                    return $this->raiseError($result);
                }
                $result = $this->query("INSERT INTO ${seqname} (id) VALUES (0)");
            } else {
                $repeat = 0;
            }
        } while ($repeat);

        if (DB::isError($result)) {
            return $this->raiseError($result);
        }
        $result = $this->query("SELECT id FROM ${seqname}");
        if (DB::isError($result)) {
            return $result;
        }
        $row = $result->fetchRow(DB_FETCHMODE_ORDERED);
        if (DB::isError($row || !$row)) {
            return $row;
        }
        return $row[0];
    }

    // }}}
    // {{{ createSequence()

    /**
     * Creates a new sequence
     *
     * @param string $seq_name  name of the new sequence
     *
     * @return int  DB_OK on success.  A DB_Error object is returned if
     *              problems arise.
     *
     * @internal
     * @see DB_common::createSequence()
     * @access public
     */
    function createSequence($seq_name)
    {
        $seqname = $this->getSequenceName($seq_name);
        return $this->query("CREATE TABLE ${seqname} ".
                            '(id INT NOT NULL UNIQUE)');
    }
    // }}}
    // {{{ dropSequence()

    /**
     * Deletes a sequence
     *
     * @param string $seq_name  name of the sequence to be deleted
     *
     * @return int  DB_OK on success.  DB_Error if problems.
     *
     * @internal
     * @see DB_common::dropSequence()
     * @access public
     */
    function dropSequence($seq_name)
    {
        $seqname = $this->getSequenceName($seq_name);
        return $this->query("DROP TABLE ${seqname}");
    }
    // }}}
    // {{{ quote()

    /**
    * Quote the given string so it can be safely used within string delimiters
    * in a query.
    * @param $string mixed Data to be quoted
    * @return mixed "NULL" string, quoted string or original data
    */
    function quote($str = null)
    {
        switch (strtolower(gettype($str))) {
            case 'null':
                return 'NULL';
            case 'integer':
            case 'double':
                return $str;
            case 'string':
            default:
                $str = str_replace("'", "''", $str);
                return "'$str'";
        }
    }

    // }}}
    // {{{ errorNative()

    /**
     * Get the native error code of the last error (if any) that
     * occured on the current connection.
     *
     * @access public
     *
     * @return int native ODBTP error code
     */
    function errorNative()
    {
        if (!isset($this->connection) || !is_resource($this->connection)) {
            return @odbtp_last_error();
        }
        return @odbtp_last_error($this->connection);
    }

    // }}}
    // {{{ odbtpRaiseError()

    /**
     * Gather information about an error, then use that info to create a
     * DB error object and finally return that object.
     *
     * @param  integer  $code  PEAR error number (usually a DB constant) if
     *                         manually raising an error
     * @return object  DB error object
     * @see errorCode()
     * @see errorNative()
     * @see DB_common::raiseError()
     */
    function odbtpRaiseError($code = null)
    {
        if ($code === null) {
            $code = $this->errorCode(@odbtp_last_error_state());
        }
        return $this->raiseError($code, null, null, null, @odbtp_last_error());
    }

    // }}}
    // {{{ tableInfo()

    /**
     * Returns information about a table or a result set.
     *
     * NOTE: only supports 'table' and 'flags' if <var>$result</var>
     * is a table name.
     *
     * @param object|string  $result  DB_result object from a query or a
     *                                string containing the name of a table
     * @param int            $mode    a valid tableInfo mode
     * @return array  an associative array with the information requested
     *                or an error object if something is wrong
     * @access public
     * @internal
     * @see DB_common::tableInfo()
     */
    function tableInfo($result, $mode = null)
    {
        if (isset($result->result)) {
            /*
             * Probably received a result object.
             * Extract the result resource identifier.
             */
            $id = $result->result;
            $is_table = false;
        } else if (is_string($result)) {
            /*
             * Probably received a table name.
             */
            $is_table = true;
        } else {
            /*
             * Probably received a result resource identifier.
             * Copy it.
             * Deprecated.  Here for compatibility only.
             */
            $id = $result;
            $is_table = false;
        }

        $res = array();

        if ($mode) {
            $res['num_fields'] = 0;
        }

        if ($this->options['portability'] & DB_PORTABILITY_LOWERCASE) {
            $case_func = 'strtolower';
        } else {
            $case_func = 'strval';
        }

        // if $result is a string, then we want information about a
        // table without a resultset

        if ($is_table) {
            $db = $this->_db;
            $id = @odbtp_query("||SQLColumns|$db||$result", $this->connection);
            if (!$id) {
                return $this->odbtpRaiseError();
            }
            for ($i=0; ($row = @odbtp_fetch_row($id)); $i++) {
                $res[$i]['table'] = $case_func($row[2]);
                $res[$i]['name']  = $case_func($row[3]);
                $res[$i]['type']  = $case_func($row[5]);
                $res[$i]['len']   = $row[6];
                $res[$i]['flags'] = @odbtp_flags($row[4], $row[5], $row[10]);
            }
            $count = $i;
            $id = @odbtp_query("||SQLPrimaryKeys|$db||$result", $this->connection);
            if ($id) {
                while ($row = @odbtp_fetch_row($id)) {
                    for ($i=0; $i<$count; $i++) {
                        if (!strcasecmp($row[3], $res[$i]['name'])) {
                            if( $res[$i]['flags'] )
                                $res[$i]['flags'] .= ' primary_key';
                            else
                                $res[$i]['flags'] .= 'primary_key';
                            break;
                        }
                    }
                }
                @odbtp_free_query($id);
            }
        } else { // else we want information about a resultset
            $count = @odbtp_num_fields($id);
            for ($i=0; $i<$count; $i++) {
                $res[$i]['table'] = $case_func(@odbtp_field_table($id, $i));
                $res[$i]['name']  = $case_func(@odbtp_field_name($id, $i));
                $res[$i]['type']  = $case_func(@odbtp_field_type($id, $i));
                $res[$i]['len']   = @odbtp_field_length($id, $i);
                $res[$i]['flags'] = @odbtp_field_flags($id, $i);
            }
        }
        if ($mode) { // full
            $res['num_fields'] = $count;

            for ($i=0; $i<$count; $i++) {
                if ($mode & DB_TABLEINFO_ORDER) {
                    $res['order'][$res[$i]['name']] = $i;
                }
                if ($mode & DB_TABLEINFO_ORDERTABLE) {
                    $res['ordertable'][$res[$i]['table']][$res[$i]['name']] = $i;
                }
            }
        }
        return $res;
    }

    // }}}
    // {{{ getSpecialQuery()

    /**
    * Returns the query needed to get some backend info
    * @param string $type What kind of info you want to retrieve
    * @return string The SQL query string
    */
    function getSpecialQuery($type)
    {
        switch ($this->odbc_driver) {
            case ODB_DRIVER_MSSQL:
            case ODB_DRIVER_SYBASE:
                switch ($type) {
                    case 'tables':
                        $sql = "select name from sysobjects where type = 'U' order by name";
                        break;
                    case 'views':
                        $sql = "select name from sysobjects where type = 'V' order by name";
                        break;
                    default:
                        return null;
                }
            case ODB_DRIVER_ORACLE:
                switch ($type) {
                    case 'tables':
                        $sql = "SELECT table_name FROM user_tables";
                        break;
                    default:
                        return null;
                }
            case ODB_DRIVER_MYSQL:
                switch ($type) {
                    case 'tables':
                        $sql = "SHOW TABLES";
                        break;
                    case 'databases':
                        $sql = "SHOW DATABASES";
                        break;
                    default:
                        return null;
                }
            default:
                return null;
        }
        return $sql;
    }

    // }}}

}

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 */

?>
